@@ -18,18 +18,31 @@ DEFAULT_LIMIT = 2 ** 20  # 1Mb
 log = logging.getLogger("sockio")
 
 
+class ConnectionEOFError(ConnectionError):
+    pass
+
+
+class ConnectionTimeoutError(ConnectionError):
+    pass
+
+
 def ensure_connection(f):
     assert asyncio.iscoroutinefunction(f)
+    name = f.__name__
 
     @functools.wraps(f)
     async def wrapper(self, *args, **kwargs):
         if self.auto_reconnect and not self.connected():
             await self.open()
-        timeout = kwargs.pop('timeout', self.timeout)
+        timeout = kwargs.pop("timeout", self.timeout)
         coro = f(self, *args, **kwargs)
         if timeout is not None:
             coro = asyncio.wait_for(coro, timeout)
-        return await coro
+        try:
+            return await coro
+        except asyncio.TimeoutError:
+            addr = str((self.host, self.port))
+            raise ConnectionTimeoutError("{} call timeout on {}".format(name, addr))
 
     return wrapper
 
@@ -44,8 +57,9 @@ def raw_handle_read(f):
         except ConnectionError:
             await self.close()
             raise
-        if not reply and self.at_eof():
+        if not reply:
             await self.close()
+            raise ConnectionEOFError("Connection closed by peer")
         return reply
 
     return wrapper
@@ -86,7 +100,7 @@ class StreamReader(asyncio.StreamReader):
             return e.partial
         except asyncio.LimitOverrunError as e:
             if self._buffer.startswith(eol, e.consumed):
-                del self._buffer[:e.consumed + len(eol)]
+                del self._buffer[: e.consumed + len(eol)]
             else:
                 self._buffer.clear()
             self._maybe_resume_transport()
@@ -129,6 +143,63 @@ async def open_connection(
     return reader, writer
 
 
+class BaseStream:
+    """Base asynchronous iterator stream helper for TCP connections"""
+
+    def __init__(self, tcp):
+        self.tcp = tcp
+
+    async def _read(self):
+        raise NotImplementedError
+
+    def __aiter__(self):
+        return self
+
+    async def __anext__(self):
+        try:
+            return await self._read()
+        except ConnectionEOFError:
+            raise StopAsyncIteration
+
+
+class LineStream(BaseStream):
+    """Line based asynchronous iterator stream helper for TCP connections"""
+
+    def __init__(self, tcp, eol=None):
+        super().__init__(tcp)
+        self.eol = eol
+
+    async def _read(self):
+        return await self.tcp.readline(eol=self.eol)
+
+
+class BlockStream(BaseStream):
+    """
+    Fixed based asynchronous iterator stream helper for TCP connections.
+
+    - If limit is an int the block is of fixed size
+      (TCP.readexactly semantics)
+    - If limit is a string, block ends when the limit string is
+      found (TCP.readuntil semantics)
+    """
+
+    def __init__(self, tcp, limit):
+        super().__init__(tcp)
+        self.limit = limit
+
+    async def _read(self):
+        try:
+            if isinstance(self.limit, int):
+                return await self.tcp.readexactly(self.limit)
+            else:
+                return await self.tcp.readuntil(self.limit)
+        except asyncio.IncompleteReadError as error:
+            if error.partial:
+                raise
+            else:
+                raise ConnectionEOFError()
+
+
 class TCP:
     def __init__(
         self,
@@ -142,7 +213,8 @@ class TCP:
         buffer_size=DEFAULT_LIMIT,
         no_delay=True,
         tos=IPTOS_LOWDELAY,
-        timeout=None
+        connection_timeout=None,
+        timeout=None,
     ):
         self.host = host
         self.port = port
@@ -155,6 +227,7 @@ class TCP:
         self.on_eof_received = on_eof_received
         self.no_delay = no_delay
         self.tos = tos
+        self.connection_timeout = connection_timeout
         self.timeout = timeout
         self.reader = None
         self.writer = None
@@ -169,17 +242,10 @@ class TCP:
                 self._log.warning("could not close stream: loop closed")
 
     def __aiter__(self):
-        return self
-
-    @ensure_connection
-    async def __anext__(self):
-        val = await self.reader.readline()
-        if val == b"":
-            raise StopAsyncIteration
-        return val
+        return LineStream(self)
 
     async def open(self, **kwargs):
-        timeout = kwargs.get('timeout', self.timeout)
+        connection_timeout = kwargs.get("timeout", self.connection_timeout)
         if self.connected():
             raise ConnectionError("socket already open")
         self._log.debug("open connection (#%d)", self.connection_counter + 1)
@@ -194,9 +260,15 @@ class TCP:
             no_delay=self.no_delay,
             tos=self.tos,
         )
-        if timeout is not None:
-            coro = asyncio.wait_for(coro, timeout)
-        self.reader, self.writer = await coro
+        if connection_timeout is not None:
+            coro = asyncio.wait_for(coro, connection_timeout)
+
+        try:
+            self.reader, self.writer = await coro
+        except asyncio.TimeoutError:
+            addr = self.host, self.port
+            raise ConnectionTimeoutError("Connect call timeout on {}".format(addr))
+
         if self.on_connection_made is not None:
             try:
                 res = self.on_connection_made()
@@ -238,17 +310,11 @@ class TCP:
     async def _readuntil(self, separator=b"\n"):
         return await self.reader.readuntil(separator)
 
+    @raw_handle_read
     async def _readline(self, eol=None):
         if eol is None:
             eol = self.eol
-        try:
-            reply = await self.reader.readline(eol=eol)
-        except ConnectionError as err:
-            await self.close()
-            raise err
-        if not reply:
-            await self.close()
-        return reply
+        return await self.reader.readline(eol=eol)
 
     async def _readlines(self, n, eol=None):
         if eol is None:
@@ -256,8 +322,6 @@ class TCP:
         replies = []
         for i in range(n):
             reply = await self._readline(eol=eol)
-            if not reply:
-                break
             replies.append(reply)
         return replies
 
@@ -331,42 +395,3 @@ class TCP:
     def reset_input_buffer(self):
         if self.connected():
             self.reader.reset()
-
-
-def parse_args(args=None):
-    import argparse
-
-    parser = argparse.ArgumentParser()
-    log_level_choices = ["critical", "error", "warning", "info", "debug"]
-    log_level_choices += [i.upper() for i in log_level_choices]
-    parser.add_argument("--host", default="0", help="SCPI device host name / IP")
-    parser.add_argument("-p", "--port", type=int, help="SCPI device port")
-    parser.add_argument(
-        "-r", "--request", default="*IDN?\n", help="SCPI request [%(default)s]"
-    )
-    parser.add_argument("--log-level", choices=log_level_choices, default="warning")
-    parser.add_argument("-d", "--debug", action="store_true")
-    options = parser.parse_args(args)
-    if not options.request.endswith("\n"):
-        options.request += "\n"
-    fmt = "%(asctime)-15s %(levelname)-5s %(threadName)s %(name)s: %(message)s"
-    logging.basicConfig(level=options.log_level.upper(), format=fmt)
-    return options
-
-
-async def run(options):
-    sock = TCP(options.host, options.port)
-    request = options.request
-    nb_lines = request.count("\n")
-    lines = await sock.write_readlines(request.encode(), nb_lines)
-    for line in lines:
-        print(line)
-
-
-def main(args=None):
-    options = parse_args(args=args)
-    return run(options)
-
-
-if __name__ == "__main__":
-    asyncio.run(main())
